""" CumulusCI Tasks for running Apex Tests """

import html
import io
import json
import os
import re
from typing import Dict, List, Optional

from cumulusci.core.config import TaskConfig
from cumulusci.core.exceptions import (
    ApexTestException,
    CumulusCIException,
    TaskOptionsError,
)
from cumulusci.core.utils import decode_to_unicode, determine_managed_mode
from cumulusci.tasks.salesforce import BaseSalesforceApiTask
from cumulusci.utils.http.requests_utils import safe_json_from_response
from cumulusci.utils.options import (
    CCIOptions,
    Field,
    ListOfStringsOption,
    MappingOption,
    PercentageOption,
)
from cumulusci.vcs.utils.list_modified_files import ListModifiedFiles

APEX_LIMITS = {
    "Soql": {
        "Label": "TESTING_LIMITS: Number of SOQL queries",
        "SYNC": 100,
        "ASYNC": 200,
    },
    "Email": {
        "Label": "TESTING_LIMITS: Number of Email Invocations",
        "SYNC": 10,
        "ASYNC": 10,
    },
    "AsyncCalls": {
        "Label": "TESTING_LIMITS: Number of future calls",
        "SYNC": 50,
        "ASYNC": 50,
    },
    "DmlRows": {
        "Label": "TESTING_LIMITS: Number of DML rows",
        "SYNC": 10000,
        "ASYNC": 10000,
    },
    "Cpu": {"Label": "TESTING_LIMITS: Maximum CPU time", "SYNC": 10000, "ASYNC": 60000},
    "QueryRows": {
        "Label": "TESTING_LIMITS: Number of query rows",
        "SYNC": 50000,
        "ASYNC": 50000,
    },
    "Dml": {
        "Label": "TESTING_LIMITS: Number of DML statements",
        "SYNC": 150,
        "ASYNC": 150,
    },
    "MobilePush": {
        "Label": "TESTING_LIMITS: Number of Mobile Apex push calls",
        "SYNC": 10,
        "ASYNC": 10,
    },
    "Sosl": {
        "Label": "TESTING_LIMITS: Number of SOSL queries",
        "SYNC": 20,
        "ASYNC": 20,
    },
    "Callouts": {
        "Label": "TESTING_LIMITS: Number of callouts",
        "SYNC": 100,
        "ASYNC": 100,
    },
}


TEST_RESULT_QUERY = """
SELECT Id,ApexClassId,TestTimestamp,
       Message,MethodName,Outcome,
       RunTime,StackTrace,
       (SELECT
          Id,Callouts,AsyncCalls,DmlRows,Email,
          LimitContext,LimitExceptions,MobilePush,
          QueryRows,Sosl,Cpu,Dml,Soql
        FROM ApexTestResults)
FROM ApexTestResult
WHERE AsyncApexJobId='{}'
"""


class MappingIntOption(MappingOption):
    """Parses a Mapping of Str->Int from a string in format a:b,c:d"""

    @classmethod
    def from_str(cls, v) -> Dict[str, int]:
        """Validate and convert a value.
        If its a string, parse it, else, just validate it.
        """
        try:
            v = {
                key: PercentageOption.from_str(value)
                for key, value in super().from_str(v).items()
            }
        except ValueError:
            raise TaskOptionsError(
                "Value should be a percentage or integer (e.g. 90% or 90)"
            )
        return v


class ListOfRegexPatternsOption(ListOfStringsOption):
    @classmethod
    def validate(cls, v):
        """Validate and convert a value.
        If its a string, parse it, else, just validate it.
        """
        regex_patterns: List[re.Pattern] = []
        for regex in v:
            try:
                regex_patterns.append(re.compile(regex))
            except re.error as e:
                raise TaskOptionsError(
                    "An invalid regular expression ({}) was provided ({})".format(
                        regex, e
                    )
                )
        v = regex_patterns
        return v


class RunApexTests(BaseSalesforceApiTask):
    """Task to run Apex tests with the Tooling API and report results.

    This task optionally supports retrying unit tests that fail due to
    transitory issues or concurrency-related row locks. To enable retries,
    add ones or more regular expressions to the list option `retry_failures`.

    When a test run fails, if all of the failures' error messages or stack traces
    match one of these regular expressions, each failed test will be retried by
    itself. This is often useful when running Apex tests in parallel; row locks
    may automatically be retried. Note that retries are supported whether or not
    the org has parallel Apex testing enabled.

    The ``retry_always`` option modifies this behavior: if a test run fails and
    any (not all) of the failures match the specified regular expressions,
    all of the failed tests will be retried in serial. This is helpful when
    underlying row locking errors are masked by custom exceptions.

    A useful base configuration for projects wishing to use retries is:

    .. code-block:: yaml

        retry_failures:
            - "unable to obtain exclusive access to this record"
            - "UNABLE_TO_LOCK_ROW"
            - "connection was cancelled here"
        retry_always: True

    Some projects' unit tests produce so many concurrency errors that
    it's faster to execute the entire run in serial mode than to use retries.
    Serial and parallel mode are configured in the scratch org definition file."""

    api_version = "38.0"
    name = "RunApexTests"

    class Options(CCIOptions):
        test_name_match: Optional[ListOfStringsOption] = Field(
            None,
            description=(
                "Pattern to find Apex test classes to run "
                '("%" is wildcard).  Defaults to '
                "project__test__name_match from project config. "
                "Comma-separated list for multiple patterns."
            ),
        )
        test_name_exclude: Optional[ListOfStringsOption] = Field(
            None,
            description=(
                "Query to find Apex test classes to exclude "
                '("%" is wildcard).  Defaults to '
                "project__test__name_exclude from project config. "
                "Comma-separated list for multiple patterns."
            ),
        )
        namespace: Optional[str] = Field(
            None,
            description=(
                "Salesforce project namespace.  Defaults to "
                "project__package__namespace"
            ),
        )
        managed: bool = Field(
            False,
            description=(
                "If True, search for tests in the namespace " "only. Defaults to False"
            ),
        )
        poll_interval: int = Field(
            1,
            description="Seconds to wait between polling for Apex test results.",
        )
        junit_output: Optional[str] = Field(
            "test_results.xml",
            description="File name for JUnit output.  Defaults to test_results.xml",
        )
        json_output: Optional[str] = Field(
            "test_results.json",
            description="File name for json output.  Defaults to test_results.json",
        )
        retry_failures: ListOfRegexPatternsOption = Field(
            [],
            description="A list of regular expression patterns to match against "
            "test failures. If failures match, the failing tests are retried in "
            "serial mode.",
        )
        retry_always: bool = Field(
            False,
            description="By default, all failures must match retry_failures to perform "
            "a retry. Set retry_always to True to retry all failed tests if any failure matches.",
        )
        required_org_code_coverage_percent: PercentageOption = Field(
            0,
            description="Require at least X percent code coverage across the org following the test run.",
        )
        required_per_class_code_coverage_percent: PercentageOption = Field(
            0,
            description="Require at least X percent code coverage for every class in the org.",
        )
        required_individual_class_code_coverage_percent: MappingIntOption = Field(
            {},
            description="Mapping of class names to their minimum coverage percentage requirements. "
            "Takes priority over required_per_class_code_coverage_percent for specified classes.",
        )
        verbose: bool = Field(
            False,
            description="By default, only failures get detailed output. "
            "Set verbose to True to see all passed test methods.",
        )
        test_suite_names: ListOfStringsOption = Field(
            [],
            description="List of test suite names. Only runs test classes that are part of the test suites specified.",
        )
        dynamic_filter: Optional[str] = Field(
            None,
            description="Defines a dynamic filter to apply to test classes from the org that match test_name_match. Supported values: "
            "'package_only' - only runs test classes that exist in the default package directory (force-app/ or src/),"
            "'delta_changes' - only runs test classes that are affected by the delta changes in the current branch (force-app/ or src/),"
            "Default is None, which means no dynamic filter is applied and all test classes from the org that match test_name_match are run."
            "Setting this option, the org code coverage will not be calculated.",
        )
        base_ref: Optional[str] = Field(
            None,
            description="Git reference (branch, tag, or commit) to compare against for delta changes. "
            "If not set, uses the default branch of the repository. Only used when dynamic_filter is 'delta_changes'.",
        )

    parsed_options: Options

    def _init_options(self, kwargs):
        super(RunApexTests, self)._init_options(kwargs)

        # Set defaults from project config
        if self.parsed_options.test_name_match is None:
            self.parsed_options.test_name_match = ListOfStringsOption.from_str(
                self.project_config.project__test__name_match
            )
        if self.parsed_options.test_name_exclude is None:
            self.parsed_options.test_name_exclude = ListOfStringsOption.from_str(
                self.project_config.project__test__name_exclude
            )
        if self.parsed_options.test_suite_names is None:
            self.parsed_options.test_suite_names = ListOfStringsOption.from_str(
                self.project_config.project__test__suite__names
            )
        if self.parsed_options.namespace is None:
            self.parsed_options.namespace = (
                self.project_config.project__package__namespace
            )

        self.verbose = self.parsed_options.verbose
        self.counts = {}

        self.code_coverage_level = (
            self.parsed_options.required_org_code_coverage_percent
        )

        self.required_per_class_code_coverage_percent = (
            self.parsed_options.required_per_class_code_coverage_percent
        )

        # Parse individual class coverage requirements
        # Validator already converted values to int, so just use it directly
        self.required_individual_class_code_coverage_percent = (
            self.parsed_options.required_individual_class_code_coverage_percent
        )

        # Raises a TaskOptionsError when the user provides both test_suite_names and test_name_match.
        if self.parsed_options.test_suite_names and not (
            any(
                pattern in ["%_TEST%", "%TEST%"]
                for pattern in self.parsed_options.test_name_match
            )
        ):
            raise TaskOptionsError(
                "Both test_suite_names and test_name_match cannot be passed simultaneously"
            )

    # pylint: disable=W0201
    def _init_class(self):
        self.classes_by_id = {}
        self.classes_by_name = {}
        self.job_id = None
        self.results_by_class_name = {}
        self.result = None
        self.retry_details = None

    def _get_namespace_filter(self):

        if self.parsed_options.managed:

            namespace = self.parsed_options.namespace

            if not namespace:
                raise TaskOptionsError(
                    "Running tests in managed mode but no namespace available."
                )
            namespace = "'{}'".format(namespace)
        elif self.org_config.namespace:
            namespace = self.org_config.namespace
            namespace = "'{}'".format(namespace)
        else:
            namespace = "null"
        return namespace

    def _get_test_class_query(self):
        namespace = self._get_namespace_filter()
        # Split by commas to allow multiple class name matching options
        included_tests = []
        for pattern in self.parsed_options.test_name_match:
            if pattern:
                included_tests.append("Name LIKE '{}'".format(pattern))
        # Add any excludes to the where clause
        excluded_tests = []
        for pattern in self.parsed_options.test_name_exclude:
            if pattern:
                excluded_tests.append("(NOT Name LIKE '{}')".format(pattern))
        # Get all test classes for namespace
        query = "SELECT Id, Name FROM ApexClass " + "WHERE NamespacePrefix = {}".format(
            namespace
        )

        if included_tests:
            query += " AND ({})".format(" OR ".join(included_tests))
        if excluded_tests:
            query += " AND {}".format(" AND ".join(excluded_tests))
        return query

    def _get_test_classes(self):
        # If test_suite_names is provided, execute only tests that are a part of the list of test suites provided.
        if self.parsed_options.test_suite_names:
            test_classes_from_test_suite_names = (
                self._get_test_classes_from_test_suite_names()
            )
            return test_classes_from_test_suite_names

        # test_suite_names is not provided. Fetch all the test classes from the org.
        else:
            return self._get_all_test_classes()

    def _get_all_test_classes(self):
        # Fetches all the test classes from the org.
        query = self._get_test_class_query()
        self.logger.info("Fetching all the test classes...")
        result = self.tooling.query_all(query)
        self.logger.info("Found {} test classes".format(result["totalSize"]))
        return result

    def _get_comma_separated_string_of_items(self, itemlist):
        # Accepts a list of strings. A formatted string is returned.
        # Example: Input: ['TestSuite1', 'TestSuite2']      Output: ''TestSuite1','TestSuite2''
        return ",".join([f"'{item}'" for item in itemlist])

    def _get_test_suite_ids_from_test_suite_names_query(self, test_suite_names_arg):
        # Returns a query string which when executed fetches the test suite ids of the list of test suite names.
        test_suite_names = self._get_comma_separated_string_of_items(
            test_suite_names_arg.split(",")
        )
        query1 = f"SELECT Id, TestSuiteName FROM ApexTestSuite WHERE TestSuiteName IN ({test_suite_names})"
        return query1

    def _get_test_classes_from_test_suite_ids_query(self, testSuiteIds):
        # Returns a query string which when executed fetches Apex test classes for the given list of test suite ids.
        # Apex test classes passed under test_name_exclude are ignored.
        testSuiteIds_formatted = self._get_comma_separated_string_of_items(testSuiteIds)

        if len(testSuiteIds_formatted) == 0:
            testSuiteIds_formatted = "''"

        condition = ""

        # Check if test_name_exclude is provided. Append to query string if the former is specified.
        if self.parsed_options.test_name_exclude:
            test_name_exclude = self._get_comma_separated_string_of_items(
                self.parsed_options.test_name_exclude
            )
            condition = f"AND Name NOT IN ({test_name_exclude})"

        query = f"SELECT Id, Name FROM ApexClass WHERE Id IN (SELECT ApexClassId FROM TestSuiteMembership WHERE ApexTestSuiteId IN ({testSuiteIds_formatted})) {condition}"
        return query

    def _get_test_classes_from_test_suite_names(self):
        # Returns a list of Apex test classes that belong to the test suite(s) specified. Test classes specified in test_name_exclude are excluded.
        query1 = self._get_test_suite_ids_from_test_suite_names_query(
            self.parsed_options.test_suite_names
        )
        self.logger.info("Fetching test suite metadata...")
        result = self.tooling.query_all(query1)
        testSuiteIds = []

        for record in result["records"]:
            testSuiteIds.append(str(record["Id"]))

        query2 = self._get_test_classes_from_test_suite_ids_query(testSuiteIds)
        self.logger.info("Fetching test classes belonging to the test suite(s)...")
        result = self.tooling.query_all(query2)
        self.logger.info("Found {} test classes".format(result["totalSize"]))
        return result

    def _class_exists_in_package(self, class_name):
        """Check if an Apex class exists in the default package directory."""
        package_path = self.project_config.default_package_path

        # Walk through the package directory to find .cls files
        for root, dirs, files in os.walk(package_path):
            for file in files:
                if file.endswith(".cls"):
                    # Extract class name from filename (remove .cls extension)
                    file_class_name = file[:-4]
                    if file_class_name == class_name:
                        return True
        return False

    def _filter_package_classes(self, test_classes):
        """Filter test classes to only include those that exist in the package directory."""
        if self.parsed_options.dynamic_filter is None:
            return test_classes

        filtered_records = []
        match self.parsed_options.dynamic_filter:
            case "package_only":
                filtered_records = self._filter_test_classes_to_package_only(
                    test_classes
                )
            case "delta_changes":
                filtered_records = self._filter_test_classes_to_delta_changes(
                    test_classes
                )
            case _:
                raise TaskOptionsError(
                    f"Unsupported dynamic filter: {self.parsed_options.dynamic_filter}"
                )

        # Update the result with filtered records
        filtered_result = {
            "totalSize": len(filtered_records),
            "records": filtered_records,
            "done": test_classes.get("done", True),
        }

        return filtered_result

    def _filter_test_classes_to_package_only(self, test_classes):
        """Filter test classes to only include those that exist in the package directory."""
        filtered_records = []
        excluded_count = 0

        for record in test_classes["records"]:
            class_name = record["Name"]
            if self._class_exists_in_package(class_name):
                filtered_records.append(record)
            else:
                excluded_count += 1
                self.logger.debug(
                    f"Excluding test class '{class_name}' - not found in package directory"
                )

        if excluded_count > 0:
            self.logger.info(
                f"Excluded {excluded_count} test class(es) not in package directory"
            )
        return filtered_records

    def _filter_test_classes_to_delta_changes(self, test_classes):
        """Filter test classes to only include those that are affected by the delta changes in the current branch."""
        # Check if the current base folder has git. (project_config.repo)
        if self.project_config.get_repo() is None:
            self.logger.info("No git repository found. Returning all test classes.")
            return test_classes["records"]

        self.logger.info("")
        self.logger.info("Getting the list of committed files in the current branch.")
        # Get the list of modified files in the current branch.
        task = ListModifiedFiles(
            self.project_config,
            TaskConfig(
                {
                    "options": {
                        "base_ref": self.parsed_options.base_ref,
                        "file_extensions": ["cls", "flow-meta.xml", "trigger"],
                        "directories": ["force-app", "src"],
                    }
                }
            ),
            org_config=None,
        )
        task()

        branch_return_values = task.return_values.copy()

        # Get the list of modified files which are not yet committed.
        self.logger.info("")
        self.logger.info("Getting the list of uncommitted files in the current branch.")
        task.parsed_options.base_ref = "HEAD"
        task()
        uncommitted_return_values = task.return_values.copy()

        # Get the list of changed files.
        branch_files = set(branch_return_values.get("files", []))
        uncommitted_files = set(uncommitted_return_values.get("files", []))
        changed_files = branch_files.union(uncommitted_files)

        if not changed_files:
            self.logger.info(
                "No changed files found in package directories (force-app/ or src/)."
            )
            return []

        # Extract class names from changed files
        affected_class_names = branch_return_values.get("file_names", set()).union(
            uncommitted_return_values.get("file_names", set())
        )

        if not affected_class_names:
            self.logger.info("No file names found in changed files.")
            return []

        self.logger.info(
            f"Found {len(affected_class_names)} affected class(es): {', '.join(sorted(affected_class_names))}"
        )

        # Filter test classes to only include those affected by the delta changes
        filtered_records = []
        excluded_count = 0

        affected_class_names_lower = [name.lower() for name in affected_class_names]

        for record in test_classes["records"]:
            test_class_name = record["Name"]
            if self._is_test_class_affected(
                test_class_name.lower(), affected_class_names_lower
            ):
                filtered_records.append(record)
            else:
                excluded_count += 1
                self.logger.debug(
                    f"Excluding test class '{test_class_name}' - not affected by delta changes"
                )

        if excluded_count > 0:
            self.logger.info(
                f"Excluded {excluded_count} test class(es) not affected by delta changes"
            )

        if not filtered_records:
            self.logger.info(
                "No test classes found that are affected by delta changes."
            )
            return []

        self.logger.info(
            f"Running {len(filtered_records)} test class(es) that are affected by delta changes. Test classes: {', '.join([record['Name'] for record in filtered_records])}"
        )

        return filtered_records

    def _is_test_class_affected(self, test_class_name, affected_class_names):
        """Check if a test class is affected by the changed classes."""
        # Direct match: test class name matches a changed class
        if test_class_name in affected_class_names:
            return True

        # Check if test class name corresponds to an affected class
        # Common patterns:
        # - Account.cls changed -> AccountTest.cls should run
        # - MyService.cls changed -> MyServiceTest.cls should run
        # - AccountHandler.cls changed -> AccountHandlerTest.cls should run
        for affected_class in affected_class_names:
            # Check if test class name follows common test naming patterns
            if (
                test_class_name == f"{affected_class}test"
                or test_class_name == f"test{affected_class}"
                or test_class_name.startswith(f"{affected_class}_")
                or test_class_name.startswith(f"test{affected_class}_")
                or test_class_name == f"{affected_class.replace('_', '')}test"
            ):
                return True

        return False

    def _get_test_methods_for_class(self, class_name):
        result = self.tooling.query(
            f"SELECT SymbolTable FROM ApexClass WHERE Name='{class_name}'"
        )
        test_methods = []

        try:
            methods = result["records"][0]["SymbolTable"]["methods"]
        except (TypeError, IndexError, KeyError):
            raise CumulusCIException(
                f"Unable to acquire symbol table for failed Apex class {class_name}"
            )
        for m in methods:
            for a in m.get("annotations", []):
                if a["name"].lower() in ["istest", "testmethod"]:
                    test_methods.append(m["name"])
                    break

        return test_methods

    def _is_retriable_error_message(self, error_message):
        return any(
            [reg.search(error_message) for reg in self.parsed_options.retry_failures]
        )

    def _is_retriable_failure(self, test_result):
        return self._is_retriable_error_message(
            test_result["Message"] or ""
        ) or self._is_retriable_error_message(test_result["StackTrace"] or "")

    def _get_test_results(self, allow_retries=True):
        # We need to query at both the test result and test queue item level.
        # Some concurrency problems manifest as all or part of the class failing,
        # without leaving behind any visible ApexTestResult records.
        # See https://salesforce.stackexchange.com/questions/262893/any-way-to-get-consistent-test-counts-when-parallel-testing-is-used

        # First, gather the Ids of failed test classes.
        test_classes = self.tooling.query_all(
            "SELECT Id, Status, ExtendedStatus, ApexClassId FROM ApexTestQueueItem "
            + "WHERE ParentJobId = '{}' AND Status = 'Failed'".format(self.job_id)
        )
        class_level_errors = {
            each_class["ApexClassId"]: each_class["ExtendedStatus"]
            for each_class in test_classes["records"]
        }

        result = self.tooling.query_all(TEST_RESULT_QUERY.format(self.job_id))

        if allow_retries:
            self.retry_details = {}

        for test_result in result["records"]:
            class_name = self.classes_by_id[test_result["ApexClassId"]]
            self.results_by_class_name[class_name][
                test_result["MethodName"]
            ] = test_result
            self.counts[test_result["Outcome"]] += 1

        # If we have class-level failures that did not come with line-level
        # failure details, report those as well.
        for class_id, error in class_level_errors.items():
            class_name = self.classes_by_id[class_id]

            self.logger.error(
                f"Class {class_name} failed to run some tests with the message {error}. Applying error to unit test results."
            )

            # In Spring '20, we cannot get symbol tables for managed classes.
            if self.parsed_options.managed:
                self.logger.error(
                    f"Cannot access symbol table for managed class {class_name}. Failure will not be retried."
                )
                continue

            # Get all the method names for this class
            test_methods = self._get_test_methods_for_class(class_name)
            for test_method in test_methods:
                # If this method was not run due to a class-level failure,
                # synthesize a failed result.
                # If we're retrying and fail again, do the same.
                if (
                    test_method not in self.results_by_class_name[class_name]
                    or self.results_by_class_name[class_name][test_method]["Outcome"]
                    == "Fail"
                ):
                    self.results_by_class_name[class_name][test_method] = {
                        "ApexClassId": class_id,
                        "MethodName": test_method,
                        "Outcome": "Fail",
                        "Message": f"Containing class {class_name} failed with message {error}",
                        "StackTrace": "",
                        "RunTime": 0,
                    }
                    self.counts["Fail"] += 1

        if allow_retries:
            for class_name, results in self.results_by_class_name.items():
                for test_result in results.values():
                    # Determine whether this failure is retriable.
                    if test_result["Outcome"] == "Fail" and allow_retries:
                        can_retry_this_failure = self._is_retriable_failure(test_result)
                        if can_retry_this_failure:
                            self.counts["Retriable"] += 1

                        # Even if this failure is not retriable per se,
                        # persist its details if we might end up retrying
                        # all failures.
                        if self.parsed_options.retry_always or can_retry_this_failure:
                            self.retry_details.setdefault(
                                test_result["ApexClassId"], []
                            ).append(test_result["MethodName"])

    def _process_test_results(self):
        test_results = []
        class_names = list(self.results_by_class_name.keys())
        class_names.sort()
        for class_name in class_names:
            self.retry_details = {}
            method_names = list(self.results_by_class_name[class_name].keys())
            # Added to process for the None methodnames

            if None in method_names:
                class_id = self.classes_by_name[class_name]
                self.retry_details.setdefault(class_id, []).append(
                    self._get_test_methods_for_class(class_name)
                )
                del self.results_by_class_name[class_name][None]
                self.logger.info(
                    f"Retrying class with id: {class_id} name:{class_name} due to `None` methodname"
                )
                self.counts["Retriable"] += len(self.retry_details[class_id])
                self._attempt_retries()

            has_failures = any(
                result["Outcome"] in ["Fail", "CompileFail"]
                for result in self.results_by_class_name[class_name].values()
            )
            if has_failures or self.verbose:
                self.logger.info(f"Class: {class_name}")
            method_names = list(self.results_by_class_name[class_name].keys())
            method_names.sort()
            for method_name in method_names:
                result = self.results_by_class_name[class_name][method_name]
                message = f"\t{result['Outcome']}: {result['MethodName']}"
                duration = result["RunTime"]
                result["stats"] = self._get_stats_from_result(result)
                if duration:
                    message += f" ({duration}ms)"
                test_results.append(
                    {
                        "Children": result.get("children", None),
                        "ClassName": decode_to_unicode(class_name),
                        "Method": decode_to_unicode(result["MethodName"]),
                        "Message": decode_to_unicode(result["Message"]),
                        "Outcome": decode_to_unicode(result["Outcome"]),
                        "StackTrace": decode_to_unicode(result["StackTrace"]),
                        "Stats": result.get("stats", None),
                        "TestTimestamp": result.get("TestTimestamp", None),
                    }
                )
                if result["Outcome"] in ["Fail", "CompileFail"]:
                    self.logger.info(message)
                    self.logger.info(f"\tMessage: {result['Message']}")
                    self.logger.info(f"\tStackTrace: {result['StackTrace']}")
                elif self.verbose:
                    self.logger.info(message)
        self.logger.info("-" * 80)
        self.logger.info(
            "Pass: {}  Retried: {}  Fail: {}  CompileFail: {}  Skip: {}".format(
                self.counts["Pass"],
                self.counts["Retriable"],
                self.counts["Fail"],
                self.counts["CompileFail"],
                self.counts["Skip"],
            )
        )
        self.logger.info("-" * 80)
        if self.counts["Fail"] or self.counts["CompileFail"]:
            self.logger.error("-" * 80)
            self.logger.error("Failing Tests")
            self.logger.error("-" * 80)
            counter = 0
            for result in test_results:
                if result["Outcome"] in ["Fail", "CompileFail"]:
                    counter += 1
                    self.logger.error(
                        "{}: {}.{} - {}".format(
                            counter,
                            result["ClassName"],
                            result["Method"],
                            result["Outcome"],
                        )
                    )
                    self.logger.error(f"\tMessage: {result['Message']}")
                    self.logger.error(f"\tStackTrace: {result['StackTrace']}")

        return test_results

    def _get_stats_from_result(self, result):
        stats = {"duration": result["RunTime"]}

        if result.get("ApexTestResults", None):
            for limit_name, details in APEX_LIMITS.items():
                limit_use = result["ApexTestResults"]["records"][0][limit_name]
                limit_allowed = details[
                    result["ApexTestResults"]["records"][0]["LimitContext"]
                ]
                stats[details["Label"]] = {"used": limit_use, "allowed": limit_allowed}

        return stats

    def _enqueue_test_run(self, class_ids):
        if isinstance(class_ids, dict):
            body = {
                "tests": [
                    {"classId": class_id, "testMethods": class_ids[class_id]}
                    for class_id in class_ids
                ]
            }
        else:
            body = {"classids": ",".join(class_ids)}

        return safe_json_from_response(
            self.tooling._call_salesforce(
                method="POST",
                url=self.tooling.base_url + "runTestsAsynchronous",
                json=body,
            )
        )

    def _init_task(self):
        super()._init_task()

        self.parsed_options.managed = determine_managed_mode(
            self.options, self.project_config, self.org_config
        )

    def _run_task(self):
        result = self._get_test_classes()

        # Apply dynamic filters if enabled
        if self.parsed_options.dynamic_filter:
            result = self._filter_package_classes(result)

        if result["totalSize"] == 0:
            return
        for test_class in result["records"]:
            self.classes_by_id[test_class["Id"]] = test_class["Name"]
            self.classes_by_name[test_class["Name"]] = test_class["Id"]
            self.results_by_class_name[test_class["Name"]] = {}
        self.logger.info("Queuing tests for execution...")

        self.counts = {
            "Pass": 0,
            "Fail": 0,
            "CompileFail": 0,
            "Skip": 0,
            "Retriable": 0,
        }
        self.job_id = self._enqueue_test_run(
            (str(id) for id in self.classes_by_id.keys())
        )

        self._wait_for_tests()
        self._get_test_results()

        # Did we get back retriable test results? Check our retry policy,
        # then enqueue new runs individually, until either (a) all retriable
        # tests succeed or (b) a test fails.
        able_to_retry = (
            self.counts["Retriable"] and self.parsed_options.retry_always
        ) or (
            self.counts["Retriable"] and self.counts["Retriable"] == self.counts["Fail"]
        )
        if not able_to_retry:
            self.counts["Retriable"] = 0
        else:
            self._attempt_retries()

        test_results = self._process_test_results()
        self._write_output(test_results)

        if self.counts.get("Fail") or self.counts.get("CompileFail"):
            raise ApexTestException(
                "{} tests failed and {} tests failed compilation".format(
                    self.counts.get("Fail"), self.counts.get("CompileFail")
                )
            )

        if self.code_coverage_level or self.required_per_class_code_coverage_percent:
            if self.parsed_options.managed:
                self.logger.info(
                    "This org contains a managed installation; not checking code coverage."
                )
            elif self.parsed_options.dynamic_filter is not None:
                self.logger.info("Dynamic filter is set; not checking code coverage.")
            else:
                self._check_code_coverage()
        else:
            self.logger.info(
                "No code coverage level specified; not checking code coverage."
            )

    def _check_code_coverage(self):
        self.logger.info("Checking code coverage.")
        class_level_coverage_failures = {}

        # Query for Class level code coverage using the aggregate
        if (
            self.required_per_class_code_coverage_percent
            or self.required_individual_class_code_coverage_percent
        ):
            test_classes = self.tooling.query(
                "SELECT ApexClassOrTrigger.Name, ApexClassOrTriggerId, NumLinesCovered, NumLinesUncovered FROM ApexCodeCoverageAggregate ORDER BY ApexClassOrTrigger.Name ASC"
            )["records"]

            coverage_percentage = 0
            for class_level in test_classes:
                class_name = class_level["ApexClassOrTrigger"]["Name"]
                total = (
                    class_level["NumLinesCovered"] + class_level["NumLinesUncovered"]
                )
                # prevent division by 0 errors
                if total:
                    # calculate coverage percentage
                    coverage_percentage = round(
                        (class_level["NumLinesCovered"] / total) * 100,
                        2,
                    )

                # Determine the required coverage for this class using fallback logic
                required_coverage = None
                if class_name in self.required_individual_class_code_coverage_percent:
                    # Individual class requirement takes priority
                    required_coverage = (
                        self.required_individual_class_code_coverage_percent[class_name]
                    )
                elif self.required_per_class_code_coverage_percent:
                    # Fall back to global per-class requirement
                    required_coverage = self.required_per_class_code_coverage_percent

                # Only check if a requirement is defined for this class
                if (
                    required_coverage is not None
                    and coverage_percentage < required_coverage
                ):
                    class_level_coverage_failures[class_name] = {
                        "actual": coverage_percentage,
                        "required": required_coverage,
                    }

        # Query for OrgWide coverage
        result = self.tooling.query("SELECT PercentCovered FROM ApexOrgWideCoverage")
        coverage = result["records"][0]["PercentCovered"]

        errors = []
        if (
            self.required_per_class_code_coverage_percent
            or self.required_individual_class_code_coverage_percent
        ):
            if class_level_coverage_failures:
                for class_name, coverage_info in class_level_coverage_failures.items():
                    errors.append(
                        f"{class_name}'s code coverage of {coverage_info['actual']}% is below required level of {coverage_info['required']}%."
                    )
            else:
                # Build a message about what requirements were met
                if (
                    self.required_per_class_code_coverage_percent
                    and self.required_individual_class_code_coverage_percent
                ):
                    self.logger.info(
                        f"All classes meet code coverage expectations (global: {self.required_per_class_code_coverage_percent}%, individual class requirements also satisfied)."
                    )
                elif self.required_per_class_code_coverage_percent:
                    self.logger.info(
                        f"All classes meet code coverage expectations of {self.required_per_class_code_coverage_percent}%."
                    )
                elif self.required_individual_class_code_coverage_percent:
                    self.logger.info(
                        "All classes with individual coverage requirements meet their expectations."
                    )

        if coverage < self.code_coverage_level:
            errors.append(
                f"Organization-wide code coverage of {coverage}% is below required level of {self.code_coverage_level}"
            )
        else:
            self.logger.info(
                f"Organization-wide code coverage of {coverage}% meets expectations."
            )

        if errors:
            error_message = "\n".join(errors)
            self.logger.info(error_message)
            raise ApexTestException(error_message)

    def _attempt_retries(self):
        total_method_retries = sum(
            [len(test_list) for test_list in self.retry_details.values()]
        )
        self.logger.warning(
            "Retrying {} failed methods from {} test classes".format(
                total_method_retries, len(self.retry_details)
            )
        )
        self.counts["Fail"] = 0

        for class_id, test_list in self.retry_details.items():
            for each_test in test_list:
                self.logger.warning(
                    "Retrying {}.{}".format(self.classes_by_id[class_id], each_test)
                )
                self.job_id = self._enqueue_test_run({class_id: [each_test]})
                self._wait_for_tests()
                self._get_test_results(allow_retries=False)

        # If the retry failed, report the remaining failures.
        if self.counts["Fail"]:
            self.logger.error("Test retry failed.")

    def _wait_for_tests(self):
        self.poll_complete = False
        self.poll_interval_s = self.parsed_options.poll_interval
        self.poll_count = 0
        self._poll()

    def _poll_action(self):
        self.result = self.tooling.query_all(
            "SELECT Id, Status, ApexClassId FROM ApexTestQueueItem "
            + "WHERE ParentJobId = '{}'".format(self.job_id)
        )
        counts = {
            "Aborted": 0,
            "Completed": 0,
            "Failed": 0,
            "Holding": 0,
            "Preparing": 0,
            "Processing": 0,
            "Queued": 0,
        }
        processing_class_id = None
        total_test_count = self.result["totalSize"]
        for test_queue_item in self.result["records"]:
            counts[test_queue_item["Status"]] += 1
            if test_queue_item["Status"] == "Processing":
                processing_class_id = test_queue_item["ApexClassId"]
        processing_class = ""
        if counts["Processing"] == 1:
            processing_class = f" ({self.classes_by_id[processing_class_id]})"
        self.logger.info(
            "Completed: {}  Processing: {}{}  Queued: {}".format(
                counts["Completed"],
                counts["Processing"],
                processing_class,
                counts["Queued"],
            )
        )
        if (
            total_test_count
            == counts["Completed"] + counts["Failed"] + counts["Aborted"]
        ):
            self.logger.info("Apex tests completed")
            self.poll_complete = True

    def _write_output(self, test_results):
        junit_output = self.parsed_options.junit_output
        if junit_output:
            with io.open(junit_output, mode="w", encoding="utf-8") as f:
                f.write('<testsuite tests="{}">\n'.format(len(test_results)))
                for result in test_results:
                    s = '  <testcase classname="{}" name="{}"'.format(
                        result["ClassName"], result["Method"]
                    )
                    if (
                        "Stats" in result
                        and result["Stats"]
                        and "duration" in result["Stats"]
                    ):
                        s += ' time="{}"'.format(result["Stats"]["duration"])
                    if result["Outcome"] in ["Fail", "CompileFail"]:
                        s += ">\n"
                        s += '    <failure type="failed" '
                        if result["Message"]:
                            s += 'message="{}"'.format(html.escape(result["Message"]))
                        s += ">"

                        if result["StackTrace"]:
                            s += "<![CDATA[{}]]>".format(
                                html.escape(result["StackTrace"])
                            )
                        s += "</failure>\n"
                        s += "  </testcase>\n"
                    else:
                        s += " />\n"
                    f.write(str(s))
                f.write("</testsuite>")

        json_output = self.parsed_options.json_output
        if json_output:
            with io.open(json_output, mode="w", encoding="utf-8") as f:
                f.write(str(json.dumps(test_results, indent=4)))



import json
import os
import pickle
from datetime import date, datetime
from logging import Logger
from typing import NamedTuple, Optional

# Delay saving as JSON for a few CumulusCI releases because
# people might downgrade a release and then their
# CCI can't parse their JSON orgfiles
#
# Thus we roll out the ability to parse JSON configs a bit
# ahead of the write behaviour.
SHOULD_SAVE_AS_JSON = os.environ.get("SHOULD_SAVE_AS_JSON", "False") != "False"


def load_config_from_json_or_pickle(b: bytes) -> dict:
    """Input should be plaintext JSON or Pickle"""
    assert isinstance(b, bytes)

    try:
        data = try_load_config_from_json_or_pickle(b)
    except pickle.PickleError as e:
        # we use ValueError because Pickle and Crypto both do too
        raise ValueError(str(e)) from e

    return data


class JSONSerializer(NamedTuple):
    type: str
    to_json: callable
    from_json: callable

    @property
    def name(self):
        return self.type.__name__


# make sure that datetime comes before date
string_serializers = [
    JSONSerializer(
        datetime,
        to_json=lambda x: x.isoformat(),
        from_json=lambda x: datetime.fromisoformat(x),
    ),
    JSONSerializer(
        date, to_json=lambda x: x.isoformat(), from_json=lambda x: date.fromisoformat(x)
    ),
    JSONSerializer(
        bytes,
        to_json=lambda x: x.decode("unicode_escape"),
        from_json=lambda x: x.encode("unicode_escape"),
    ),
]


def encode_value(x):
    """Encode a value that JSON does not support natively"""
    for serializer in string_serializers:
        if isinstance(x, serializer.type):
            return {"$type": serializer.name, "$value": serializer.to_json(x)}

    raise TypeError(type(x))  # pragma: no cover


def decode_dict(x: dict):
    """Decode a dict from JSON"""
    assert isinstance(x, dict)
    if "$type" in x:
        return decode_typed_value(x)
    else:
        return x


def decode_typed_value(x: dict):
    """Decode a value that JSON does not support natively"""
    for serializer in string_serializers:
        if x["$type"] == serializer.name:
            return serializer.from_json(x["$value"])

    raise TypeError(f"Unknown $type: {x['$type']}")  # pragma: no cover


def try_load_config_from_json_or_pickle(data: bytes) -> dict:
    """Load JSON or Pickle into a dict"""
    try:
        config = json.loads(data.decode("utf-8"), object_hook=decode_dict)
        # remove this debugging tool after transition
        config["serialization_format"] = "json"
        return config
    except ValueError as e1:
        try:
            # first byte in a Pickle must be part of
            # OPCODE Proto == \x80 == 128
            # https://github.com/python/cpython/blob/1b293b60067f6f4a95984d064ce0f6b6d34c1216/Lib/pickletools.py#L2124
            if data[0] != 128:
                raise ValueError("Decryption error")
            config = pickle.loads(data, encoding="bytes")
            # remove this debugging tool after transition
            config["serialization_format"] = "pickle"
            return config
        except pickle.UnpicklingError as e2:
            raise ValueError(f"{e1}\n{e2}")


def report_error(msg: str, e: Exception, logger: Logger):
    logger.error(
        "\n".join(
            (
                msg,
                str(e),
                "Please report it to the CumulusCI team.",
                "For now this is just a warning. By January 2023 it may become a real error.",
                "When you report it to the CumulusCI team, they will investigate",
                "whether the error is in your config or in CumulusCI",
            )
        )
    )


def check_round_trip(data: dict, logger: Logger) -> Optional[bytes]:
    """Return JSON bytes if possible, else None"""
    try:
        as_json_text = json.dumps(data, default=encode_value).encode("utf-8")
    except Exception as e:
        report_error("CumulusCI found an unusual datatype in your config:", e, logger)
        return None
    try:
        test_load = load_config_from_json_or_pickle(as_json_text)
        assert _simplify_config(test_load) == _simplify_config(
            data
        ), f"JSON did not round-trip-cleanly {test_load}, {data}"
    except Exception as e:  # pragma: no cover
        report_error("CumulusCI found a problem saving your config:", e, logger)
        return None
    assert isinstance(as_json_text, bytes)
    return as_json_text


def _simplify_config(config: dict):
    return {k: v for k, v in config.items() if k != "serialization_format"}


def serialize_config_to_json_or_pickle(config: dict, logger: Logger) -> bytes:
    """Serialize a dict to JSON if possible or Pickle otherwise"""
    as_json_text = check_round_trip(config, logger)
    if as_json_text and SHOULD_SAVE_AS_JSON:
        assert isinstance(as_json_text, bytes)
        return as_json_text
    else:
        return pickle.dumps(config, protocol=2)



import collections
import logging
import tempfile
import typing as T
from contextlib import contextmanager, nullcontext
from pathlib import Path

from requests.structures import CaseInsensitiveDict as RequestsCaseInsensitiveDict
from simple_salesforce import Salesforce
from sqlalchemy import Boolean, Column, MetaData, Table, Unicode, inspect
from sqlalchemy.engine.base import Connection
from sqlalchemy.orm import Session, mapper

from cumulusci.core.enums import StrEnum
from cumulusci.core.exceptions import BulkDataException
from cumulusci.utils.iterators import iterate_in_chunks


class DataApi(StrEnum):
    """Enum defining requested Salesforce data API for an operation."""

    BULK = "bulk"
    REST = "rest"
    SMART = "smart"


class CaseInsensitiveDict(RequestsCaseInsensitiveDict):
    def __init__(self, *args, **kwargs):
        self._canonical_keys = {}
        super().__init__(*args, **kwargs)

    def canonical_key(self, name):
        return self._canonical_keys[name.lower()]

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        self._canonical_keys[key.lower()] = key


class SqlAlchemyMixin:
    logger: logging.Logger
    metadata: MetaData
    models: dict
    options: dict
    session: Session
    sf: Salesforce

    def _create_record_type_table(self, table_name):
        """Create a table to store mapping between Record Type Ids and Developer Names."""
        rt_map_model_name = f"{table_name}Model"
        self.models[table_name] = type(rt_map_model_name, (object,), {})
        rt_map_fields = [
            Column("record_type_id", Unicode(18), primary_key=True),
            Column("developer_name", Unicode(255)),
            Column("is_person_type", Boolean),
        ]
        rt_map_table = Table(table_name, self.metadata, *rt_map_fields)
        mapper(self.models[table_name], rt_map_table)

    def _extract_record_types(
        self, sobject, tablename: str, conn, is_person_accounts_enabled: bool
    ):
        """Query for Record Type information and persist it in the database."""
        self.logger.info(f"Extracting Record Types for {sobject}")
        if is_person_accounts_enabled:
            query = f"SELECT Id, DeveloperName, IsPersonType FROM RecordType WHERE SObjectType='{sobject}'"
        else:
            query = f"SELECT Id, DeveloperName FROM RecordType WHERE SObjectType='{sobject}'"

        result = self.sf.query(query)

        if result["totalSize"]:
            sql_bulk_insert_from_records(
                connection=conn,
                table=self.metadata.tables[tablename],
                columns=["record_type_id", "developer_name", "is_person_type"],
                record_iterable=(
                    [rt["Id"], rt["DeveloperName"], rt.get("IsPersonType", False)]
                    for rt in result["records"]
                ),
            )

    @contextmanager
    def _temp_database_url(self):
        with tempfile.TemporaryDirectory() as t:
            tempdb = Path(t) / "temp_db.db"

            self.logger.debug(f"Using temporary database {tempdb}")
            database_url = f"sqlite:///{tempdb}"
            yield database_url

    def _database_url(self):
        database_url = self.options.get("database_url")
        if database_url:
            return nullcontext(enter_result=database_url)
        else:
            return self._temp_database_url()


def _handle_primary_key(mapping, fields):
    """Provide support for legacy mappings which used the OID as the pk but
    default to using an autoincrementing int pk and a separate sf_id column"""

    id_column = "id"
    if mapping.get_oid_as_pk():
        # Get Id column from mapping
        id_column = mapping.fields["Id"]

    fields.append(Column(id_column, Unicode(255), primary_key=True))


def create_table(mapping, metadata) -> Table:
    """Given a mapping data structure (from mapping.yml) and SQLAlchemy
    metadata, create a table matching the mapping.

    Mapping should be a MappingStep instance"""

    fields = []
    _handle_primary_key(mapping, fields)

    # make a field list to create
    for field, db in mapping.get_complete_field_map().items():
        if field == "Id":
            continue

        fields.append(Column(db, Unicode(255)))

    if mapping.record_type:
        fields.append(Column("record_type", Unicode(255)))
    return create_table_if_needed(mapping.table, metadata, fields)


def create_table_if_needed(tablename, metadata, fields: T.List[Column]) -> Table:
    t = Table(tablename, metadata, *fields)
    inspector = inspect(metadata.bind)
    if inspector.has_table(tablename):
        raise BulkDataException(f"Table already exists: {tablename}")
    t.create(metadata.bind)
    return t


def generate_batches(num_records, batch_size):
    """Generate batch size list for splitting a number of tasks into batch jobs.

    Given a number of records to split up, and a batch size, generate a
    stream of batchsize, index pairs"""
    num_batches, left_over = divmod(num_records, batch_size)
    if left_over:
        num_batches += 1  # need an extra batch for the clean-up
    for i in range(0, num_batches):
        is_last_batch = i == num_batches - 1
        if is_last_batch and left_over:
            batch_size = left_over
        if batch_size > 0:
            yield batch_size, i, num_batches


class RowErrorChecker:
    def __init__(self, logger, ignore_row_errors, row_warning_limit):
        self.logger = logger
        self.ignore_row_errors = ignore_row_errors
        self.row_warning_limit = row_warning_limit
        self.row_error_count = 0

    def check_for_row_error(self, result, row_id):
        if not result.success:
            msg = f"Error on record with id {row_id}: {result.error}"
            if self.ignore_row_errors:
                if self.row_error_count < self.row_warning_limit:
                    self.logger.warning(msg)
                elif self.row_error_count == self.row_warning_limit:
                    self.logger.warning("Further warnings suppressed")
                self.row_error_count += 1
                return self.row_error_count
            else:
                raise BulkDataException(msg)


def consume(iterator):
    """Consume an iterator for its side effects.

    Simplified from the function in https://docs.python.org/3/library/itertools.html
    """
    collections.deque(iterator, maxlen=0)


def sql_bulk_insert_from_records(
    *,
    connection: Connection,
    table: Table,
    columns: T.Tuple[str],
    record_iterable: T.Iterable,
) -> None:
    """Persist records from the given generator into the local database."""
    consume(
        sql_bulk_insert_from_records_incremental(
            connection=connection,
            table=table,
            columns=columns,
            record_iterable=record_iterable,
        )
    )


def sql_bulk_insert_from_records_incremental(
    *,
    connection: Connection,
    table: Table,
    columns: T.Tuple[str],
    record_iterable: T.Iterable,
):
    """Generator that persists batches of records from the given generator into the local database

    Yields after every batch."""
    dict_iterable = (dict(zip(columns, row)) for row in record_iterable)
    for group in iterate_in_chunks(10000, dict_iterable):
        with connection.begin():
            connection.execute(table.insert(), group)
        # self.session.flush()  -- Did this line do anything?
        yield


def sf_query_to_table(
    *,
    table: Table,
    metadata: MetaData,
    connection: Connection,
    sobject: str,
    fields: T.List[str],
    **queryargs,
) -> Table:
    """Cache data from Salesforce in a SQL Alchemy Table."""
    from cumulusci.tasks.bulkdata.step import DataOperationStatus, get_query_operation

    qs = get_query_operation(sobject=sobject, fields=fields, **queryargs)
    qs.query()
    if qs.job_result.status is not DataOperationStatus.SUCCESS:  # pragma: no cover
        raise BulkDataException(
            f"Unable to query records for {sobject}: {','.join(qs.job_result.job_errors)}"
        )
    results = qs.get_results()

    sql_bulk_insert_from_records(
        connection=connection,
        table=table,
        columns=tuple(fields),
        record_iterable=results,
    )
    return table



from io import StringIO
from typing import Optional

from cumulusci.core.config.tests.test_config import DummyRepository
from cumulusci.core.dependencies.base import DynamicDependency
from cumulusci.vcs.base import VCSService
from cumulusci.vcs.models import (
    AbstractBranch,
    AbstractComparison,
    AbstractGitTag,
    AbstractPullRequest,
    AbstractRef,
    AbstractRepo,
    AbstractRepoCommit,
)
from cumulusci.vcs.utils import AbstractCommitDir


class ConcreteVCSService(VCSService):
    service_type = "test_service"

    @classmethod
    def validate_service(cls, options, keychain):
        return {"validated": True}

    def get_repository(self, options: dict = {}):
        return DummyRepo()

    def parse_repo_url(self):
        pass

    @classmethod
    def get_service_for_url(cls, config, url, service_alias=None):
        return None

    @property
    def dynamic_dependency_class(self) -> DynamicDependency:
        return DynamicDependency

    def get_committer(self, repo: AbstractRepo) -> AbstractCommitDir:
        return DummyCommitDir()

    def markdown(self):
        pass

    def parent_pr_notes_generator(self, repo):
        pass

    def release_notes_generator(self, options: dict):
        pass


class DummyCommitDir(AbstractCommitDir):
    def __call__(
        self, local_dir, branch, repo_dir=None, commit_message=None, dry_run=False
    ):
        pass


class DummyTag(AbstractGitTag):
    def __init__(self, tag, **kwargs):
        super().__init__(**kwargs)
        self.tag = tag
        self.sha = "1234567890abcdef"

    @property
    def message(self) -> str:
        return "Dummy tag message"


class DummyRef(AbstractRef):
    pass


class DummyBranch(AbstractBranch):
    def branches(self):
        pass

    def get_branch(self):
        pass

    def commit(self):
        pass


class DummyRepo(AbstractRepo):

    repo: DummyRepository

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.repo = kwargs.get("repo", None)

    def create_tag(
        self, tag_name: str, message: str, sha: str, obj_type: str, tagger={}
    ) -> "AbstractGitTag":
        return DummyTag(tag_name)

    def get_ref_for_tag(self, tag_name: str) -> "AbstractRef":
        return DummyRef(f"ref-{tag_name}")

    def get_tag_by_ref(
        self, ref: AbstractRef, tag_name: str = None
    ) -> "AbstractGitTag":
        return DummyTag(f"tag-{ref}-{tag_name}")

    def branch(self, branch_name: str):
        pass

    def branches(self):
        pass

    def compare_commits(self, base: str, head: str, source: str):
        pass

    def merge(self, base: str, head: str, source: str, message: str = ""):
        pass

    def archive(self, format: str, zip_content, ref=None):
        pass

    def create_pull(
        self,
        title: str,
        base: str,
        head: str,
        body: str = None,
        maintainer_can_modify: bool = None,
        options: dict = {},
    ):
        pass

    def create_release(
        self,
        tag_name: str,
        name: str,
        body: str = None,
        draft: bool = False,
        prerelease: bool = False,
        options: dict = {},
    ):
        pass

    @property
    def default_branch(self):
        pass

    def full_name(self):
        pass

    def get_commit(self, commit_sha: str):
        pass

    def pull_requests(self, **kwargs):
        pass

    def release_from_tag(self, tag_name: str):
        pass

    def releases(self):
        pass

    def get_pr_issue_labels(self, pull_request):
        pass

    def has_issues(self):
        pass

    def latest_release(self):
        pass

    @property
    def owner_login(self):
        pass

    def directory_contents(self, subfolder: str, return_as, ref: str):
        pass

    @property
    def clone_url(self):
        return "https://github.com/test/repo.git"

    def file_contents(self, file_path: str, ref: str = None) -> StringIO:
        return self.repo.file_contents(file_path, ref=ref)

    def get_latest_prerelease(self):
        pass

    def get_ref(self, ref_sha: str):
        pass

    def create_commit_status(
        self,
        commit_id: str,
        context: str,
        state: str,
        description: str,
        target_url: str,
    ):
        pass


class DummyComparison(AbstractComparison):
    def compare(self):
        pass

    def get_comparison(self):
        pass


class DummyRepoCommit(AbstractRepoCommit):
    pass


class DummyPullRequest(AbstractPullRequest):
    def create_pull(self):
        pass

    @property
    def number(self):
        pass

    @property
    def title(self) -> str:
        return "Dummy pull request title"

    def pull_requests(self):
        pass

    @property
    def base_ref(self) -> Optional[str]:
        super().base_ref()

    @property
    def head_ref(self) -> str:
        return "Dummy pull request title"

    @property
    def merged_at(self) -> str:
        return "Dummy pull request title"



import json
import os

import requests

from cumulusci.core.config import TaskConfig
from cumulusci.core.exceptions import CumulusCIException
from cumulusci.core.utils import process_list_arg
from cumulusci.salesforce_api.metadata import ApiListMetadata
from cumulusci.tasks.salesforce import (
    BaseRetrieveMetadata,
    BaseSalesforceApiTask,
    DescribeMetadataTypes,
)
from cumulusci.tasks.salesforce.sourcetracking import ListChanges, retrieve_components

nl = "\n"


class ListNonSourceTrackable(BaseSalesforceApiTask):

    task_options = {
        "api_version": {
            "description": "Override the API version used to list metadatatypes",
        },
    }

    def _init_task(self):
        super()._init_task()
        if "api_version" not in self.options:
            self.options[
                "api_version"
            ] = self.project_config.project__package__api_version

    def get_types_details(self, api_version):
        # The Metadata coverage report: https://developer.salesforce.com/docs/metadata-coverage/{version} is created from
        # the below URL. (So the api versions are allowed based on those report ranges)
        url = f"https://dx-extended-coverage.my.salesforce-sites.com/services/apexrest/report?version={api_version}"
        response = requests.get(url)
        if response.status_code == 200:
            json_response = response.json()
            return json_response
        else:
            raise CumulusCIException(
                f"Failed to retrieve response with status code {response.status_code}"
            )

    def _run_task(self):
        metadatatypes_details = self.get_types_details(self.options["api_version"])[
            "types"
        ]
        all_nonsource_types = []
        for md_type, details in metadatatypes_details.items():
            if not details["channels"]:
                raise CumulusCIException(
                    f"Api version {self.options['api_version']} not supported"
                )
            if (
                details["channels"]["sourceTracking"] is False
                and details["channels"]["metadataApi"] is True
            ):
                all_nonsource_types.append(md_type)

        types_supported = DescribeMetadataTypes(
            org_config=self.org_config,
            project_config=self.project_config,
            task_config=self.task_config,
        )._run_task()

        self.return_values = []
        for md_type in all_nonsource_types:
            if md_type in types_supported:
                self.return_values.append(md_type)

        if self.return_values:
            self.return_values.sort()

        self.logger.info(
            f"Non source trackable Metadata types supported by org: \n{self.return_values}"
        )
        return self.return_values


class ListComponents(BaseSalesforceApiTask):
    api_class = ApiListMetadata
    task_options = {
        "api_version": {
            "description": "Override the API version used to list metadatatypes",
        },
        "metadata_types": {"description": "A comma-separated list of metadata types."},
    }

    def _init_task(self):
        super()._init_task()

    def _init_options(self, kwargs):
        super(ListComponents, self)._init_options(kwargs)
        if "api_version" not in self.options:
            self.options[
                "api_version"
            ] = self.project_config.project__package__api_version
        if "metadata_types" not in self.options:
            self.options["metadata_types"] = ListNonSourceTrackable(
                org_config=self.org_config,
                project_config=self.project_config,
                task_config=TaskConfig(
                    {"options": {"api_version": self.options["api_version"]}}
                ),
            )._run_task()
        else:
            self.options["metadata_types"] = process_list_arg(
                self.options.get("metadata_types")
            )

    def _get_components(self):
        list_components = []
        for md_type in self.options["metadata_types"]:
            api_object = self.api_class(
                self, metadata_type=md_type, as_of_version=self.options["api_version"]
            )
            components = api_object()
            for temp in components[md_type]:
                cmp = {
                    "MemberType": md_type,
                    "MemberName": temp["fullName"],
                    "lastModifiedByName": temp["lastModifiedByName"],
                    "lastModifiedDate": temp["lastModifiedDate"],
                }
                if cmp not in list_components:
                    list_components.append(cmp)
        return list_components

    def _run_task(self):
        self.return_values = self._get_components()
        self.logger.info(
            f"Found {len(self.return_values)} non source trackable components in the org for the given types."
        )
        for change in self.return_values:
            self.logger.info("{MemberType}: {MemberName}".format(**change))
        return self.return_values


retrieve_components_task_options = ListComponents.task_options.copy()
retrieve_components_task_options["path"] = {
    "description": "The path to write the retrieved metadata",
    "required": False,
}
retrieve_components_task_options["include"] = {
    "description": "Components will be included if one of these names"
    "is part of either the metadata type or name. "
    "Example: ``-o include CustomField,Admin`` matches both "
    "``CustomField: Favorite_Color__c`` and ``Profile: Admin``"
}
retrieve_components_task_options["exclude"] = {
    "description": "Exclude components matching this name."
}
retrieve_components_task_options[
    "namespace_tokenize"
] = BaseRetrieveMetadata.task_options["namespace_tokenize"]


class RetrieveComponents(ListComponents, BaseSalesforceApiTask):
    task_options = retrieve_components_task_options

    def _init_options(self, kwargs):
        super(RetrieveComponents, self)._init_options(kwargs)
        self.options["include"] = process_list_arg(self.options.get("include", []))
        self.options["exclude"] = process_list_arg(self.options.get("exclude", []))
        self._include = self.options["include"]
        self._exclude = self.options["exclude"]
        self._exclude.extend(self.project_config.project__source__ignore or [])

        package_directories = []
        default_package_directory = None
        if os.path.exists("sfdx-project.json"):
            with open("sfdx-project.json", "r", encoding="utf-8") as f:
                sfdx_project = json.load(f)
                for package_directory in sfdx_project.get("packageDirectories", []):
                    package_directories.append(package_directory["path"])
                    if package_directory.get("default"):
                        default_package_directory = package_directory["path"]

        path = self.options.get("path")
        if path is None:
            # set default path to src for mdapi format,
            # or the default package directory from sfdx-project.json for dx format
            if (
                default_package_directory
                and self.project_config.project__source_format == "sfdx"
            ):
                path = default_package_directory
                md_format = False
            else:
                path = "src"
                md_format = True
        else:
            md_format = path not in package_directories
        self.md_format = md_format
        self.options["path"] = path

    def _run_task(self):
        components = self._get_components()
        filtered, ignored = ListChanges._filter_changes(self, components)
        if not filtered:
            self.logger.info("No components to retrieve")
            return
        for cmp in filtered:
            self.logger.info("{MemberType}: {MemberName}".format(**cmp))

        target = os.path.realpath(self.options["path"])
        package_xml_opts = {}
        if self.options["path"] == "src":
            package_xml_opts.update(
                {
                    "package_name": self.project_config.project__package__name,
                    "install_class": self.project_config.project__package__install_class,
                    "uninstall_class": self.project_config.project__package__uninstall_class,
                }
            )
        retrieve_components(
            filtered,
            self.org_config,
            target,
            md_format=self.md_format,
            namespace_tokenize=self.options.get("namespace_tokenize"),
            api_version=self.options["api_version"],
            extra_package_xml_opts=package_xml_opts,
        )



import pytest

from cumulusci.salesforce_api.org_schema_models import SObject


class TestOrgSchemaModels:
    def test_sobject__eq__true(self):
        a = SObject(name="foo", createable=False)
        b = SObject(name="foo", createable=False)
        assert a == b

    def test_sobject__eq__false(self):
        a = SObject(name="foo", createable=False)
        b = SObject(name="bar", createable=False)
        assert a != b
        b = SObject(name="bar", createable=True)
        assert a != b

    def test_repr(self):
        a = SObject(name="foo", createable=False)
        r = repr(a)
        assert "foo" in r
        assert "createable" in r

    def test_roundtrip_mappings(self, temp_db):
        with temp_db() as (connection, metadata, session):
            session.add(SObject(name="Foo", urls={"a": "b"}))
            session.commit()
        with temp_db() as (connection, metadata, session):
            foo = session.query(SObject).one()
            assert foo["name"] == "Foo"
            assert foo["urls"] == {"a": "b"}

    def test_roundtrip_mappings__empty(self, temp_db):
        # with temp_db() as open_db:
        with temp_db() as (connection, metadata, session):
            session.add(SObject(name="Foo", urls={}))
            session.commit()
        with temp_db() as (connection, metadata, session):
            foo = session.query(SObject).one()
            assert foo["name"] == "Foo"
            assert foo["urls"] == {}
            assert session.execute("select urls from sobjects").first()["urls"] is None

    def test_roundtrip_sequences(self, temp_db):
        # with temp_db() as open_db:
        with temp_db() as (connection, metadata, session):
            session.add(
                SObject(name="Foo", childRelationships=[("a", "b"), ("c", "d")])
            )
            session.commit()
        with temp_db() as (connection, metadata, session):
            foo = session.query(SObject).one()
            assert foo["name"] == "Foo"
            assert foo["childRelationships"] == (("a", "b"), ("c", "d"))

    def test_roundtrip_sequences__empty(self, temp_db):
        with temp_db() as (connection, metadata, session):
            session.add(SObject(name="Foo", childRelationships=[]))
            session.commit()
        with temp_db() as (connection, metadata, session):
            foo = session.query(SObject).one()
            assert foo["name"] == "Foo"
            assert foo["childRelationships"] == ()
            assert (
                session.execute("select childRelationships from sobjects").first()[
                    "childRelationships"
                ]
                is None
            )

    def test_getattr_getitem(self):
        so = SObject(name="Foo", childRelationships=[])
        assert so.name == so["name"]
        with pytest.raises(KeyError):
            so["aaa"]

        with pytest.raises(AttributeError):
            so.aaa



from unittest import mock

import pytest
import responses

from cumulusci.core.exceptions import BulkDataException, TaskOptionsError
from cumulusci.tasks.bulkdata import DeleteData
from cumulusci.tasks.bulkdata.step import (
    DataApi,
    DataOperationJobResult,
    DataOperationResult,
    DataOperationStatus,
    DataOperationType,
)
from cumulusci.tasks.bulkdata.tests.utils import _make_task
from cumulusci.tests.util import mock_describe_calls


class TestDeleteData:
    @responses.activate
    @mock.patch("cumulusci.tasks.bulkdata.delete.get_query_operation")
    @mock.patch("cumulusci.tasks.bulkdata.delete.get_dml_operation")
    def test_run(self, dml_mock, query_mock):
        mock_describe_calls()
        task = _make_task(DeleteData, {"options": {"objects": "Contact"}})
        query_mock.return_value.get_results.return_value = iter(
            ["001000000000000", "001000000000001"]
        )
        query_mock.return_value.job_result = DataOperationJobResult(
            DataOperationStatus.SUCCESS, [], 2, 0
        )
        dml_mock.return_value.get_results.return_value = iter(
            [
                DataOperationResult("001000000000000", True, None),
                DataOperationResult("001000000000001", True, None),
            ]
        )
        dml_mock.return_value.job_result = DataOperationJobResult(
            DataOperationStatus.SUCCESS, [], 0, 0
        )

        task()

        query_mock.assert_called_once_with(
            sobject="Contact",
            fields=["Id"],
            api_options={},
            context=task,
            query="SELECT Id FROM Contact",
            api=DataApi.SMART,
        )
        query_mock.return_value.query.assert_called_once()
        query_mock.return_value.get_results.assert_called_once()

        dml_mock.assert_called_once_with(
            sobject="Contact",
            operation=DataOperationType.DELETE,
            api_options={},
            context=task,
            fields=["Id"],
            api=DataApi.SMART,
            volume=2,
        )
        dml_mock.return_value.start.assert_called_once()
        dml_mock.return_value.end.assert_called_once()
        dml_mock.return_value.load_records.assert_called_once()
        dml_mock.return_value.get_results.assert_called_once()

    @responses.activate
    @mock.patch("cumulusci.tasks.bulkdata.delete.get_query_operation")
    @mock.patch("cumulusci.tasks.bulkdata.delete.get_dml_operation")
    def test_run__no_results(self, dml_mock, query_mock):
        mock_describe_calls()
        task = _make_task(DeleteData, {"options": {"objects": "Contact"}})
        query_mock.return_value.get_results.return_value = iter([])
        query_mock.return_value.job_result = DataOperationJobResult(
            DataOperationStatus.SUCCESS, [], 0, 0
        )

        task()

        query_mock.assert_called_once_with(
            sobject="Contact",
            fields=["Id"],
            api_options={},
            context=task,
            query="SELECT Id FROM Contact",
            api=DataApi.SMART,
        )
        query_mock.return_value.query.assert_called_once()
        query_mock.return_value.get_results.assert_not_called()

        dml_mock.assert_not_called()

    @responses.activate
    @mock.patch("cumulusci.tasks.bulkdata.delete.get_query_operation")
    @mock.patch("cumulusci.tasks.bulkdata.delete.get_dml_operation")
    def test_run__job_error_delete(self, dml_mock, query_mock):
        mock_describe_calls()
        task = _make_task(DeleteData, {"options": {"objects": "Contact"}})
        query_mock.return_value.get_results.return_value = iter(
            ["001000000000000", "001000000000001"]
        )
        query_mock.return_value.job_result = DataOperationJobResult(
            DataOperationStatus.SUCCESS, [], 2, 0
        )
        dml_mock.return_value.get_results.return_value = iter(
            [
                DataOperationResult("001000000000000", True, None),
                DataOperationResult("001000000000001", False, None),
            ]
        )
        with pytest.raises(BulkDataException):
            task()

    @responses.activate
    @mock.patch("cumulusci.tasks.bulkdata.delete.get_query_operation")
    @mock.patch("cumulusci.tasks.bulkdata.delete.get_dml_operation")
    def test_run__job_error_query(self, dml_mock, query_mock):
        mock_describe_calls()
        task = _make_task(DeleteData, {"options": {"objects": "Contact"}})
        query_mock.return_value.get_results.return_value = iter(
            ["001000000000000", "001000000000001"]
        )
        query_mock.return_value.job_result = DataOperationJobResult(
            DataOperationStatus.JOB_FAILURE, [], 0, 0
        )
        with pytest.raises(BulkDataException):
            task()

    @responses.activate
    @mock.patch("cumulusci.tasks.bulkdata.delete.get_query_operation")
    @mock.patch("cumulusci.tasks.bulkdata.delete.get_dml_operation")
    def test_run__row_error(self, dml_mock, query_mock):
        mock_describe_calls()
        task = _make_task(DeleteData, {"options": {"objects": "Contact"}})
        query_mock.return_value.get_results.return_value = iter(
            ["001000000000000", "001000000000001"]
        )
        query_mock.return_value.job_result = DataOperationJobResult(
            DataOperationStatus.SUCCESS, [], 2, 0
        )
        dml_mock.return_value.get_results.return_value = iter(
            [
                DataOperationResult("001000000000000", True, None),
                DataOperationResult("001000000000001", False, None),
            ]
        )
        with pytest.raises(BulkDataException):
            task()

    @responses.activate
    @mock.patch("cumulusci.tasks.bulkdata.delete.get_query_operation")
    @mock.patch("cumulusci.tasks.bulkdata.delete.get_dml_operation")
    def test_run__ignore_error(self, dml_mock, query_mock):
        mock_describe_calls()
        task = _make_task(
            DeleteData,
            {
                "options": {
                    "objects": "Contact",
                    "ignore_row_errors": "true",
                    "hardDelete": "true",
                }
            },
        )
        query_mock.return_value.get_results.return_value = iter(
            ["001000000000000", "001000000000001"]
        )
        query_mock.return_value.job_result = DataOperationJobResult(
            DataOperationStatus.SUCCESS, [], 2, 0
        )
        dml_mock.return_value.get_results.return_value = iter(
            [
                DataOperationResult("001000000000000", True, None),
                DataOperationResult("001000000000001", False, None),
            ]
        )
        dml_mock.return_value.job_result = DataOperationJobResult(
            DataOperationStatus.SUCCESS, [], 2, 0
        )
        with mock.patch.object(task.logger, "warning") as warning:
            task()
        assert len(warning.mock_calls) == 1
        query_mock.assert_called_once_with(
            sobject="Contact",
            api_options={},
            context=task,
            query="SELECT Id FROM Contact",
            api=DataApi.SMART,
            fields=["Id"],
        )
        query_mock.return_value.query.assert_called_once()
        query_mock.return_value.get_results.assert_called_once()

        dml_mock.assert_called_once_with(
            sobject="Contact",
            operation=DataOperationType.HARD_DELETE,
            api_options={},
            context=task,
            fields=["Id"],
            api=DataApi.SMART,
            volume=2,
        )
        dml_mock.return_value.start.assert_called_once()
        dml_mock.return_value.end.assert_called_once()
        dml_mock.return_value.load_records.assert_called_once()
        dml_mock.return_value.get_results.assert_called_once()

    @responses.activate
    @mock.patch("cumulusci.tasks.bulkdata.delete.get_query_operation")
    @mock.patch("cumulusci.tasks.bulkdata.delete.get_dml_operation")
    def test_run__ignore_error_throttling(self, dml_mock, query_mock):
        mock_describe_calls()
        task = _make_task(
            DeleteData,
            {
                "options": {
                    "objects": "Contact",
                    "ignore_row_errors": "true",
                    "hardDelete": "true",
                }
            },
        )
        query_mock.return_value.get_results.return_value = iter(
            ["001000000000000", "001000000000001"] * 15
        )
        query_mock.return_value.job_result = DataOperationJobResult(
            DataOperationStatus.SUCCESS, [], 2, 0
        )
        dml_mock.return_value.get_results.return_value = iter(
            [
                DataOperationResult("001000000000000", True, None),
                DataOperationResult("001000000000001", False, None),
            ]
            * 15
        )
        dml_mock.return_value.job_result = DataOperationJobResult(
            DataOperationStatus.SUCCESS, [], 2, 0
        )
        with mock.patch.object(task.logger, "warning") as warning:
            task()
        assert len(warning.mock_calls) == task.row_warning_limit + 1 == 11
        assert "warnings suppressed" in str(warning.mock_calls[-1])

    @responses.activate
    @mock.patch("cumulusci.tasks.bulkdata.delete.get_query_operation")
    @mock.patch("cumulusci.tasks.bulkdata.delete.get_dml_operation")
    def test_run__where(self, dml_mock, query_mock):
        mock_describe_calls()
        task = _make_task(
            DeleteData, {"options": {"objects": "Contact", "where": "Id != null"}}
        )
        query_mock.return_value.get_results.return_value = iter(
            ["001000000000000", "001000000000001"]
        )
        query_mock.return_value.job_result = DataOperationJobResult(
            DataOperationStatus.SUCCESS, [], 2, 0
        )
        dml_mock.return_value.get_results.return_value = iter(
            [
                DataOperationResult("001000000000000", True, None),
                DataOperationResult("001000000000001", True, None),
            ]
        )
        dml_mock.return_value.job_result = DataOperationJobResult(
            DataOperationStatus.SUCCESS, [], 2, 0
        )
        task()
        query_mock.assert_called_once_with(
            sobject="Contact",
            api_options={},
            context=task,
            query="SELECT Id FROM Contact WHERE Id != null",
            fields=["Id"],
            api=DataApi.SMART,
        )
        query_mock.return_value.query.assert_called_once()
        query_mock.return_value.get_results.assert_called_once()

        dml_mock.assert_called_once_with(
            sobject="Contact",
            operation=DataOperationType.DELETE,
            api_options={},
            context=task,
            fields=["Id"],
            api=DataApi.SMART,
            volume=2,
        )
        dml_mock.return_value.start.assert_called_once()
        dml_mock.return_value.end.assert_called_once()
        dml_mock.return_value.load_records.assert_called_once()
        dml_mock.return_value.get_results.assert_called_once()

    @responses.activate
    @mock.patch("cumulusci.tasks.bulkdata.delete.get_query_operation")
    @mock.patch("cumulusci.tasks.bulkdata.delete.get_dml_operation")
    def test_run__query_fails(self, dml_mock, query_mock):
        mock_describe_calls()
        task = _make_task(
            DeleteData, {"options": {"objects": "Contact", "where": "Id != null"}}
        )
        query_mock.return_value.get_results.return_value = iter(
            ["001000000000000", "001000000000001"]
        )
        query_mock.return_value.job_result = DataOperationJobResult(
            DataOperationStatus.JOB_FAILURE, [], 0, 0
        )
        with pytest.raises(BulkDataException):
            task()

    @responses.activate
    def test_validate_and_inject_namespace__standard(self):
        mock_describe_calls()

        task = _make_task(DeleteData, {"options": {"objects": "Contact,Account"}})

        task._validate_and_inject_namespace()

        assert task.sobjects == ["Contact", "Account"]

    @responses.activate
    def test_validate_and_inject_namespace__missing_object(self):
        mock_describe_calls()
        task = _make_task(
            DeleteData, {"options": {"objects": "ApexTestQueueItem,Account"}}
        )
        # ApexTestQueueItem is not deletable

        with pytest.raises(TaskOptionsError):
            task._validate_and_inject_namespace()

    def test_validate_and_inject_namespace__packaged(self):
        task = _make_task(DeleteData, {"options": {"objects": "Contact,Test__c"}})
        task.project_config.project__package__namespace = "ns"
        task.org_config = mock.Mock()
        task.org_config.salesforce_client.describe.return_value = {
            "sobjects": [
                {"name": "ns__Test__c", "deletable": True},
                {"name": "Contact", "deletable": True},
            ]
        }

        task._validate_and_inject_namespace()

        assert task.sobjects == ["Contact", "ns__Test__c"]

    def test_validate_and_inject_namespace__packaged_and_not(self):
        task = _make_task(DeleteData, {"options": {"objects": "Contact,Test__c"}})
        task.project_config.project__package__namespace = "ns"
        task.org_config = mock.Mock()
        task.org_config.salesforce_client.describe.return_value = {
            "sobjects": [
                {"name": "Test__c", "deletable": True},
                {"name": "Contact", "deletable": True},
                {"name": "ns__Test__c", "deletable": True},
            ]
        }

        task._validate_and_inject_namespace()

        # Prefer the user entry where there is ambiguity.
        assert task.sobjects == ["Contact", "Test__c"]

    def test_object_description(self):
        t = _make_task(DeleteData, {"options": {"objects": "a", "where": "Id != null"}})
        assert t._object_description("a") == 'a objects matching "Id != null"'

        t = _make_task(DeleteData, {"options": {"objects": "a"}})
        assert t._object_description("a") == "all a objects"

    def test_init_options(self):
        with pytest.raises(TaskOptionsError):
            _make_task(DeleteData, {"options": {"objects": ""}})

        with pytest.raises(TaskOptionsError):
            _make_task(DeleteData, {"options": {"objects": "a,b", "where": "x='y'"}})

        with pytest.raises(TaskOptionsError):
            _make_task(DeleteData, {"options": {"objects": "a", "api": "blah"}})

        with pytest.raises(TaskOptionsError):
            _make_task(
                DeleteData,
                {"options": {"objects": "a", "api": "rest", "hardDelete": True}},
            )

        t = _make_task(
            DeleteData,
            {
                "options": {
                    "objects": "a",
                    "where": "Id != null",
                    "hardDelete": "true",
                    "ignore_row_errors": "false",
                }
            },
        )
        assert t.options["where"] == "Id != null"
        assert not t.options["ignore_row_errors"]
        assert t.options["hardDelete"]

        t = _make_task(DeleteData, {"options": {"objects": "a,b"}})
        assert t.options["objects"] == ["a", "b"]



from io import StringIO
from unittest.mock import patch

import pytest

from cumulusci.core.exceptions import YAMLParseException
from cumulusci.utils.yaml.safer_loader import _replace_nbsp, load_yaml_data


@pytest.fixture
def incorrect_yaml_file(tmp_path):
    with (tmp_path / "foo.yml").open("w+") as cci_yml_file:
        invalid_yml = """xyz: abc   \n>>>\nefg: lmn\n"""
        cci_yml_file.write(invalid_yml)
        cci_yml_file.seek(0)
        yield cci_yml_file


def test_simple_load(caplog):
    yaml = """xyz:
        y: abc"""
    cciyml = load_yaml_data(StringIO(yaml))
    assert not caplog.text

    assert isinstance(cciyml, dict)  # should parse despite funny character
    assert cciyml["xyz"]["y"] == "abc", cciyml


def test_convert_nbsp(caplog):
    yaml = """xyz:
        \u00A0 y: abc"""
    cciyml = load_yaml_data(StringIO(yaml))
    assert "space character" in caplog.text

    assert isinstance(cciyml, dict)  # should parse despite funny character
    assert cciyml["xyz"]["y"] == "abc", cciyml


def test_converter():
    inp = """xyz:
        \u00A0 y: abc"""
    outp = """xyz:
          y: abc"""

    rc = _replace_nbsp(inp, "filename")
    assert rc == outp


def test_converter_is_selective():
    inp = """xyz:
            y: abc\u00A0"""

    rc = _replace_nbsp(inp, "filename")
    assert rc == inp


def test_invalid_cumulusci_yml_file(incorrect_yaml_file):
    with pytest.raises(
        YAMLParseException,
        match="An error occurred parsing yaml file at line 2, column 1.*",
    ):
        load_yaml_data(incorrect_yaml_file)


def test_invalid_string_io():
    invalid_yml_string = """xyz: abc   \n>>>\nefg: lmn\n"""
    with pytest.raises(YAMLParseException) as error:
        load_yaml_data(StringIO(invalid_yml_string))

    assert "Error message: " in error.value.args[0]


@patch("cumulusci.utils.yaml.safer_loader.yaml.safe_load")
def test_generic_exception__with_name_attr(safe_load, incorrect_yaml_file):
    safe_load.side_effect = Exception("generic")
    with pytest.raises(YAMLParseException):
        load_yaml_data(incorrect_yaml_file)


@patch("cumulusci.utils.yaml.safer_loader.yaml.safe_load")
def test_generic_exception__without_name_attr(safe_load):
    invalid_yaml = """xyz: abc   \n>>>\nefg: lmn\n"""
    safe_load.side_effect = Exception("generic")
    with pytest.raises(
        YAMLParseException,
        match="An error occurred parsing a yaml file.\nError message: generic",
    ):
        load_yaml_data(StringIO(invalid_yaml))



import fnmatch
import os
import re


class Base:
    """Base class for keywords, testcases, and suites

    This is mainly used to store the data passed to use from the
    listener methods.
    """

    def __init__(self, name, attrs):
        self.name = name
        self.attrs = attrs

    @property
    def longname(self):
        # Robot's a bit inconsistent here: suites and test cases have
        # a longname attribute, for the keyword the name is the longname.
        if "longname" in self.attrs:
            return self.attrs["longname"]
        else:
            return self.name


class Testcase(Base):
    def __repr__(self):
        return "<Testcase: {}>".format(self.name)


class Keyword(Base):
    def __repr__(self):
        return "<Keyword: {}>".format("  ".join([self.name] + self.attrs["args"]))


class Suite(Base):
    def __repr__(self):
        path = self.attrs["source"]
        rel_path = os.path.relpath(path)
        return "<Suite: {} ({})>".format(self.name, rel_path)


class Breakpoint:
    """A breakpoint

    :param breakpoint_type:  class of object to break on;
                             currently only Keyword is supported.
                             Eventually I want to support setting breakpoints
                             when a suite or test case is first started, before
                             any keywords run.
    :param pattern:          glob-style pattern to match against
                             the current step
    :param temporary:        if True, the breakpoint is removed once
                             it triggers. Used for setting temporary
                             breakpoints for stepping through the code

    Note: when a breakpoint is encountered, a string will be formed
    with the fully qualified name of the current testcase and the
    fully qualified name of the current keyword, separated with "::"

    The breakpoint pattern will be matched against this string.

    Example:
        Root_suite.sub_suite.a test case::BuiltIn.log

    To break on the "Salesforce.breakpoint" keyword in any test, the
    breakpoint could use any one of these patterns depending on how
    precise you want to be:

        - ``*::cumulusci.robotframework.Salesforce.Breakpoint``
        - ``*::*.Breakpoint``
        - ``*.Breakpoint``

    """

    def __init__(self, breakpoint_type, pattern, temporary=False):
        self.breakpoint_type = breakpoint_type
        self.pattern = pattern
        self.regex = fnmatch.translate(pattern)
        self.temporary = temporary

    def match(self, context):
        """Return True if the breakpoint matches the current context"""
        if re.match(self.regex, context, re.IGNORECASE):
            return True
        return False



import os
from pathlib import Path
from unittest import mock

import click
import pytest

from cumulusci.cli.tests.utils import recursive_list_files, run_click_command
from cumulusci.core.dependencies.dependencies import PackageNamespaceVersionDependency
from cumulusci.core.exceptions import NotInProject
from cumulusci.utils import temporary_dir

from .. import project
from ..runtime import CliRuntime


class TestProjectCommands:
    def test_validate_project_name(self):
        with pytest.raises(click.UsageError):
            project.validate_project_name("with spaces")

    def test_validate_project_name__valid(self):
        assert project.validate_project_name("valid") == "valid"

    @mock.patch("cumulusci.cli.project.click")
    def test_project_init(self, click):
        with temporary_dir():
            os.mkdir(".git")
            Path(".git", "HEAD").write_text("ref: refs/heads/main")

            click.prompt.side_effect = (
                "testproj",  # project_name
                "testpkg",  # package_name
                "testns",  # package_namespace
                "43.0",  # api_version
                "mdapi",  # source_format
                "3",  # extend other URL
                "https://github.com/SalesforceFoundation/NPSP",  # github_url
                "main",  # git_default_branch
                "work/",  # git_prefix_feature
                "uat/",  # git_prefix_beta
                "rel/",  # git_prefix_release
                "%_TEST%",  # test_name_match
                "90",  # code_coverage
            )
            click.confirm.side_effect = (
                True,
                True,
                True,
            )  # is managed? extending? enforce Apex coverage?

            runtime = CliRuntime(
                config={"project": {"test": {"name_match": "%_TEST%"}}},
                load_keychain=False,
            )
            run_click_command(project.project_init, runtime=runtime)

            # Make sure expected files/dirs were created
            assert [
                ".git/",
                ".git/HEAD",
                ".github/",
                ".github/PULL_REQUEST_TEMPLATE.md",
                ".gitignore",
                "README.md",
                "cumulusci.yml",
                "datasets/",
                "datasets/mapping.yml",
                "orgs/",
                "orgs/beta.json",
                "orgs/dev.json",
                "orgs/feature.json",
                "orgs/release.json",
                "robot/",
                "robot/testproj/",
                "robot/testproj/doc/",
                "robot/testproj/resources/",
                "robot/testproj/tests/",
                "robot/testproj/tests/create_contact.robot",
                "sfdx-project.json",
                "src/",
            ] == recursive_list_files()

    @mock.patch("cumulusci.cli.project.click")
    def test_project_init_tasks(self, click):
        """Verify that the generated cumulusci.yml file is readable and has the proper robot task"""
        with temporary_dir():
            os.mkdir(".git")
            Path(".git", "HEAD").write_text("ref: refs/heads/main")

            click.prompt.side_effect = (
                "testproj",  # project_name
                "testpkg",  # package_name
                "testns",  # package_namespace
                "43.0",  # api_version
                "mdapi",  # source_format
                "3",  # extend other URL
                "https://github.com/SalesforceFoundation/NPSP",  # github_url
                "main",  # git_default_branch
                "work/",  # git_prefix_feature
                "uat/",  # git_prefix_beta
                "rel/",  # git_prefix_release
                "%_TEST%",  # test_name_match
                "90",  # code_coverage
            )
            click.confirm.side_effect = (
                True,
                True,
                True,
            )  # is managed? extending? enforce code coverage?

            run_click_command(project.project_init)

            # verify we can load the generated yml
            cli_runtime = CliRuntime(load_keychain=False)

            # ...and verify it has the expected tasks
            config = cli_runtime.project_config.config_project
            expected_tasks = {
                "robot": {
                    "options": {
                        "suites": "robot/testproj/tests",
                        "options": {"outputdir": "robot/testproj/results"},
                    }
                },
                "robot_testdoc": {
                    "options": {
                        "path": "robot/testproj/tests",
                        "output": "robot/testproj/doc/testproj_tests.html",
                    }
                },
                "run_tests": {"options": {"required_org_code_coverage_percent": 90}},
            }
            assert config["tasks"] == expected_tasks

    def test_project_init_no_git(self):
        with temporary_dir():
            with pytest.raises(click.ClickException):
                run_click_command(project.project_init)

    def test_project_init_already_initted(self):
        with temporary_dir():
            os.mkdir(".git")
            Path(".git", "HEAD").write_text("ref: refs/heads/main")
            with open("cumulusci.yml", "w"):
                pass  # create empty file

            with pytest.raises(click.ClickException):
                run_click_command(project.project_init)

    def test_project_init_dont_overwrite(self):
        with temporary_dir():
            # Gotta have a Repo
            os.mkdir(".git")
            Path(".git", "HEAD").write_text("ref: refs/heads/main")

            os.mkdir("orgs")
            orgs = "orgs/"
            text = "Can't touch this"

            path_list = [
                Path("README.md"),
                Path(".gitignore"),
                Path(orgs + "dev.json"),
                Path(orgs + "release.json"),
            ]
            for path in path_list:
                path.write_text(text)

            runtime = mock.Mock()
            runtime.project_config.project = {"test": "test"}

            run_click_command(project.project_info, runtime=runtime)

            # Project init must not overwrite project files or org defs
            for path in path_list:
                assert text == path.read_text()

    @mock.patch("click.echo")
    def test_project_info(self, echo):
        runtime = mock.Mock()
        runtime.project_config.project = {"test": "test"}

        run_click_command(project.project_info, runtime=runtime)

        echo.assert_called_once_with("\x1b[1mtest:\x1b[0m test")

    def test_project_info__outside_project(self):
        runtime = mock.Mock()
        runtime.project_config = None
        runtime.project_config_error = NotInProject()
        with temporary_dir():
            with pytest.raises(NotInProject):
                run_click_command(project.project_info, runtime=runtime)

    @mock.patch("cumulusci.cli.project.get_static_dependencies")
    def test_project_dependencies(self, get_static_dependencies):
        out = []
        runtime = mock.Mock()
        runtime.project_config.project__dependencies = [
            {"namespace": "npe01", "version": "3.16"},
            {"namespace": "npsp", "version": "3.193"},
        ]
        get_static_dependencies.return_value = [
            PackageNamespaceVersionDependency(namespace="npe01", version="3.16"),
            PackageNamespaceVersionDependency(namespace="npsp", version="3.193"),
        ]

        with mock.patch("click.echo", out.append):
            run_click_command(
                project.project_dependencies,
                runtime=runtime,
                resolution_strategy="production",
            )

        assert out == [
            str(PackageNamespaceVersionDependency(namespace="npe01", version="3.16")),
            str(PackageNamespaceVersionDependency(namespace="npsp", version="3.193")),
        ]

    def test_render_recursive(self):
        out = []
        with mock.patch("click.echo", out.append):
            project.render_recursive(
                {"test": [{"list": ["list"], "dict": {"key": "value"}, "str": "str"}]}
            )
        assert """\x1b[1mtest:\x1b[0m
    -
        \x1b[1mlist:\x1b[0m
            - list
        \x1b[1mdict:\x1b[0m
            \x1b[1mkey:\x1b[0m value
        \x1b[1mstr:\x1b[0m str""" == "\n".join(
            out
        )
